﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.Threading;
using System.Threading.Tasks;
using System.Diagnostics;

public class LockingQueue<T> {
	private readonly int segmentSize;

	private Segment head;
	private Segment tail;
	private int lastSlot;

	private object thisLock = new object();

	public LockingQueue(int segmentSize) {
		this.segmentSize = segmentSize;

		head = new Segment(segmentSize, 0);

		lastSlot = -1;
		tail = head;
	}

	public void Enqueue(T item) {
		Segment myTail;
		int mySlotLocal;

		lock (thisLock) {
			myTail = tail;

			lastSlot++;

			int targetSegmentId = lastSlot / segmentSize;

			mySlotLocal = lastSlot % segmentSize;
			if (mySlotLocal == segmentSize - 1) {
				var newSegment = new Segment(segmentSize, targetSegmentId + 1);
				tail.Next = newSegment;
				tail = newSegment;
			}
		}

		myTail.Values[mySlotLocal] = item;
	}

	private class Segment {
		public Segment Next;
		public readonly int Id;
		public readonly T[] Values;

		public Segment(int valueCount, int id) {
			Id = id;
			Values = new T[valueCount];
			Next = null;
		}
	}
}

public class LockFreeQueue<T> {
	private readonly int segmentSize;

	private volatile Segment head;
	private volatile Segment tail;
	private int lastSlot;

	public LockFreeQueue(int segmentSize) {
		this.segmentSize = segmentSize;

		head = new Segment(segmentSize, 0);

		lastSlot = -1;
		tail = head;
	}

	public void Enqueue(T item) {
		Segment myTail = tail;
		int mySlot = Interlocked.Increment(ref lastSlot);

		var spinWait = new SpinWait();
		int targetSegmentId = mySlot / segmentSize;
		while (myTail.Id != targetSegmentId) {
			while (myTail.Next == null) {
				spinWait.SpinOnce();	// Will Yield for us if spinning to long.

				Interlocked.Increment(ref debugSpinCount);
			}

			myTail = myTail.Next;
		}

		int mySlotLocal = mySlot % segmentSize;
		if (mySlotLocal == segmentSize - 1) {
			var newSegment = new Segment(segmentSize, targetSegmentId + 1);
			
			// Just for debugging to see some more thread interleaving:
			// Thread.Yield();

			tail = newSegment;
			myTail.Next = newSegment;
		}

		myTail.Values[mySlotLocal] = item;
	}

	private class Segment {
		public volatile Segment Next;
		public readonly int Id;
		public readonly T[] Values;

		public Segment(int valueCount, int id) {
			Id = id;
			Values = new T[valueCount];
			Next = null;	// volatile => forces all previous writes
		}
	}

	private int debugSpinCount = 0;

	public void DebugPrintStatistics() {
		Console.WriteLine("  Items count: {0:N0}", lastSlot + 1);
		Console.WriteLine("  Spin count:  {0:N0}", debugSpinCount);
	}
}

class Program {
	static int[] SegmentSizes = { 8, 16, 128, 1024 };

	static void Main(string[] args) {
		LockFreeQueue<int> queue = null;
		LockingQueue<int> lockingQueue = null;

		int upperBound = 10000000;
		int delta = upperBound;

		for (int run = 0; run < SegmentSizes.Length; run++) {
			int segmentSize = SegmentSizes[run];
		
			queue = new LockFreeQueue<int>(segmentSize);
			lockingQueue = new LockingQueue<int>(segmentSize);

			var lockFreeStopwatch = Stopwatch.StartNew();
			Parallel.For(0, upperBound, item => {
				queue.Enqueue(item);
				queue.Enqueue(item + delta);
				queue.Enqueue(item + 2 * delta);
				queue.Enqueue(item + 3 * delta);
			});
			lockFreeStopwatch.Stop();

			var lockingStopwatch = Stopwatch.StartNew();
			Parallel.For(0, upperBound, item => {
				lockingQueue.Enqueue(item);
				lockingQueue.Enqueue(item + delta);
				lockingQueue.Enqueue(item + 2 * delta);
				lockingQueue.Enqueue(item + 3 * delta);
			});
			lockingStopwatch.Stop();

			Console.WriteLine("SegmentSize == {0}, Items == {1:N0}", segmentSize, upperBound);
			Console.WriteLine("+++ LockFreeQueue +++");
			queue.DebugPrintStatistics();
			Console.WriteLine("  Execution time: {0} ms ({1:F2} times faster than Locking)", lockFreeStopwatch.Elapsed.TotalMilliseconds, lockingStopwatch.Elapsed.TotalMilliseconds / lockFreeStopwatch.Elapsed.TotalMilliseconds);
			Console.WriteLine("--- LockingQueue ---");
			Console.WriteLine("  Execution time: {0} ms", lockingStopwatch.Elapsed.TotalMilliseconds);
			Console.WriteLine();
		}
	}
}
